/* Handle array memory issues
**
*/
// Note:
// These functions play some tricks with pointers that probably aren't technically
// kosher.  However, I think the way I have written it, the code will always work.
// The arrays are built using (real **) pointers for dimensions > 1,
// and (void *) pointers for the final dimension.

#include <mdarrays.h>
#include <internal.h>

// For passing find and line in custom_data void*
struct fill_foreach_passer {
    char *cfile;
    uint cline;
    void *value_pointer;
};
typedef struct fill_foreach_passer fill_foreach_passer_struct;

// ----------------------------------------------------
// - - - - ALLOCATION - - - -
// ----------------------------------------------------

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
void * _mdarrays_allocate_recurse(char *cfile, uint cline, uint dimensions, uint d_counter, 
                                  uint type_size, int *sizes)
{
void *return_pointer = NULL;
void *array = NULL;
char *char_pointer = NULL;
char **char_pointer2 = NULL;
uint loop;
uint current_dimension_size;
uint allocate_size;
uint allocate_extra_size;

current_dimension_size = sizes[d_counter-1];
APD_PUSH_FUNCTION(4, "dimensions=%d, d_counter=%d, type_size=%d, current size=%d",
      dimensions, d_counter, type_size, current_dimension_size);

MDARRAYS_ASSERT(dimensions > 0, "Fatal error: dimensions (%d < 1)", dimensions);
MDARRAYS_ASSERT(d_counter <= dimensions);
MDARRAYS_ASSERT(type_size > 0);
MDARRAYS_ASSERT(sizes != NULL);

// Get dimension size for this dimension
MDARRAYS_ASSERT(current_dimension_size > 0,
           "negative or zero dimension size (dimension %d)", d_counter);

// Make top level special with embedded dimensions and size info
if (d_counter == 1)
  {
    allocate_extra_size = sizeof(mdarray_info_struct);
  } // end 1-D array
else
  {
    allocate_extra_size = 0;
  } // end >1 D array

// Get regular allocation size
if (d_counter == dimensions)
  {
    allocate_size = current_dimension_size*type_size;
  }
else
  {
    allocate_size = current_dimension_size*sizeof(void *);
  }
APD_REPORT(5, "allocate_extra_size=%d, allocate_size=%d", allocate_extra_size, allocate_size);

size_t total_malloc = allocate_size + allocate_extra_size;
MDARRAYS_ASSERT(total_malloc > 0, "total_malloc=%d, allocate_size=%d, allocate_extra_size=%d",
                total_malloc, allocate_size, allocate_extra_size);
array = (void *)malloc(total_malloc);
APD_REPORT(5, "allocated %d bytes at %lX", total_malloc, (unsigned long int)array);

// Move pointer past extra information if first dimension
if (d_counter == 1)
  {
    APD_REPORT(5, "original main pointer @%lX", (unsigned long int)array);
    char_pointer = (char *)array;
    char_pointer = char_pointer + sizeof(mdarray_info_struct);
    array = (void *)char_pointer;
    APD_REPORT(5, "new main pointer @%lX", (unsigned long int)array);
  } // end 1-D array

if (d_counter == dimensions)
  {
      return_pointer = array;
  } // end if 1 dimension left
else
  {
    char_pointer2 = (char **)array;  // Must use a real type to use indexing in for loop below
    for (loop = 0; loop < current_dimension_size; loop++)
      {
        char_pointer2[loop] 
            = (char *)_mdarrays_allocate_recurse(cfile, cline,
                                                 dimensions, d_counter + 1, type_size, sizes);
        APD_REPORT(5, "char_pointer2(@%lX)[%d](@%lX)=%lX",
                   (unsigned long int)char_pointer2, loop,
                   (unsigned long int)(char_pointer2+loop),
                   (unsigned long int)char_pointer2[loop]);
        MDARRAYS_ASSERT(char_pointer2[loop] != NULL,
                        "null pointer returned from _mdarrays_allocate_recurse()"
                        "char_pointer2(@%lX)[%d](@%lX)=%lX",
                        (unsigned long int)char_pointer2, loop,
                        (unsigned long int)(char_pointer2+loop),
                        (unsigned long int)char_pointer2[loop]);
      }
    return_pointer = (void *)char_pointer2;
  } // end if d_counter < dimensions

APD_POP_FUNCTION();
return return_pointer;
}

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// This main routine just basically sets up the arguments list to pass to
// _mdarrays_allocate_recurse().
void * _mdarrays_allocate(char *cfile, uint cline,
                          uint dimensions, uint type, uint type_size, 
                          int *sizes, char *name)
{
void * array;
char *char_pointer;
mdarray_info_struct *mdarray_info;

APD_PUSH_FUNCTION(4, "dimensions=%d, type_size=%d", dimensions, type_size);

MDARRAYS_ASSERT(dimensions > 0, "Fatal error: dimensions (%d < 1)", dimensions);
MDARRAYS_ASSERT(dimensions < MDARRAYS_MAX_DIMENSIONS);
MDARRAYS_ASSERT(type_size > 0);
MDARRAYS_ASSERT(sizes != NULL);
MDARRAYS_ASSERT(sizeof(char) == 1);

APD_REPORT(5, "getting storage");
array = _mdarrays_allocate_recurse(cfile, cline, dimensions, 1, type_size, sizes);

// Get mdarray_info pointer for this array
char_pointer = (char *)array;
char_pointer = char_pointer - sizeof(mdarray_info_struct);
mdarray_info = (mdarray_info_struct *)char_pointer;
APD_REPORT(5, "mdarray_info @ %lX", (unsigned long int)mdarray_info);

// Add array info structure
mdarray_info->tag = ARRAY_INFO_STRUCT_TAG;
MDARRAYS_ASSERT(strlen(name) < MDARRAYS_ARRAY_NAME_SIZE-2, "name too long (%s)", name);
strcpy(mdarray_info->name, name);
APD_REPORT(5, "mdarray_info->name=%s", mdarray_info->name);
mdarray_info->dimensions = dimensions;
APD_REPORT(5, "mdarray_info->dimensions=%d", mdarray_info->dimensions);
mdarray_info->type_size = type_size;
APD_REPORT(5, "mdarray_info->type_size=%d", mdarray_info->type_size);
mdarray_info->type = type;
APD_REPORT(5, "mdarray_info->type=%d", mdarray_info->type);
mdarray_info->elements = 1;
for (uint index = 0; index < dimensions; index++)
  {
    mdarray_info->sizes[index] = sizes[index];
    APD_REPORT(5, "mdarray_info->sizes[%d]=%d", index, mdarray_info->sizes[index]);
    mdarray_info->elements *= sizes[index];
  }
APD_REPORT(5, "mdarray_info->elements=%d", mdarray_info->elements);
// Set default print settings based on type
_mdarrays_set_style_guts(cfile, cline,
                         MDARRAYS_PRINT_STYLE, array, true/*defaults*/, 0, 0, 0, 0, 0);
_mdarrays_set_style_guts(cfile, cline, 
                         MDARRAYS_SAVE_STYLE, array, true/*defaults*/, 0, 0, 0, 0, 0);
APD_POP_FUNCTION();
return array;
} // end _mdarrays_allocate()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// Allocate an array (this is called by _mdarrays_allocate_macro_nofill)
// +++++++++++++++++++++++++++++++++++++++++++++++++
void * _mdarrays_allocate_macro_guts(char *cfile, uint cline, char *type_string, 
                                     uint type_size, bool is_object,
                                     va_list variable_parameters)
{
int size[MDARRAYS_MAX_DIMENSIONS];
uint dimension = 0;
void *return_me;

APD_PUSH_FUNCTION(4);
MDARRAYS_ASSERT(type_size > 0);
size[dimension] = va_arg(variable_parameters, int);
APD_REPORT(5, "size[0]=%d", size[0]);
MDARRAYS_ASSERT(size[dimension] >= 0, "size[%d] < 0", dimension);
while (size[dimension] > 0)
  {
    dimension++;
    size[dimension] = va_arg(variable_parameters, int);
    APD_REPORT(5, "size[%d]=%d", dimension, size[dimension]);
    MDARRAYS_ASSERT(size[dimension] >= 0, "size[%d] < 0", dimension);
    MDARRAYS_ASSERT(size[dimension] !=  INT_MAX, "size[%d]= INT_MAX", dimension);
  }
uint type;
if (is_object)
  {
    type = MDARRAYS_GPOINTER_TYPE;
  }
else
  {
    _mdarrays_get_type_from_string(cfile, cline, type_string, &type);
  }
return_me = _mdarrays_allocate(cfile, cline, dimension, type, type_size, size, "no name");
APD_POP_FUNCTION();
return return_me;
} // end _mdarrays_allocate_macro_guts()


// +++++++++++++++++++++++++++++++++++++++++++++++++
// Allocate an array (this is called by MDARRAYS_ALLOCATE() defined in mdarrays.h)
// +++++++++++++++++++++++++++++++++++++++++++++++++
void * _mdarrays_allocate_macro_nofill(char *cfile, uint cline, 
                                       char *type_string, uint type_size, 
                                       bool is_object, ...)
{
va_list variable_parameters;
void *return_me;

APD_PUSH_FUNCTION(4, "from \"%s\", line %d", cfile, cline);
va_start(variable_parameters, is_object);
return_me = _mdarrays_allocate_macro_guts(cfile, cline, type_string, type_size, 
                                          is_object, variable_parameters);
va_end(variable_parameters);
APD_POP_FUNCTION();
return return_me;
} // end _mdarrays_allocate_macro_nofill()


// +++++++++++++++++++++++++++++++++++++++++++++++++
//
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _macro_fill_foreach_function(void *array, uint *access_index, void *user_data)
{
mdarray_info_struct *mdarray_info;
fill_foreach_passer_struct *ffps = user_data;

mdarray_info = _mdarrays_get_mdarray_info(ffps->cfile, ffps->cline, array, "unknown name");
_assign_pointer(ffps->cfile, ffps->cline, array, mdarray_info, "unknown name", 
                access_index, ffps->value_pointer);
}

// +++++++++++++++++++++++++++++++++++++++++++++++++
// Allocate an array and fill it with a value
// (this is called by MDARRAYS_ALLOCATE_AND_FILL() defined in mdarrays.h)
// +++++++++++++++++++++++++++++++++++++++++++++++++
void * _mdarrays_allocate_macro_fill(char *cfile, uint cline, char *type_string, 
                                     uint type_size, void *value_pointer, bool is_object, ...)
{
va_list variable_parameters;
fill_foreach_passer_struct ffps;
void *return_me;

APD_PUSH_FUNCTION(4, "from \"%s\", line %d", cfile, cline);

va_start(variable_parameters, is_object);
return_me = _mdarrays_allocate_macro_guts(cfile, cline, type_string, type_size, is_object,
                                          variable_parameters);
va_end(variable_parameters);

// Now do fill
ffps.cfile = cfile;
ffps.cline = cline;
ffps.value_pointer = value_pointer;
_mdarrays_foreach(cfile, cline, return_me, _macro_fill_foreach_function, &ffps); 

APD_POP_FUNCTION();
return return_me;
} // end _mdarrays_allocate_macro_fill()

// +++++++++++++++++++++++++++++++++++++++++++++++++
// Allocate an array and fill it with 0

// +++++++++++++++++++++++++++++++++++++++++++++++++
void * _mdarrays_allocate_macro_fill_zero(char *cfile, uint cline, char *type_string, 
                                          uint type_size, bool is_object, ...)
{
va_list variable_parameters;
void *return_me;
fill_foreach_passer_struct ffps;
ffps.cfile = cfile;
ffps.cline = cline;

unsigned char zero_unsigned_char = 0;
char zero_char = 0;
uint zero_unsigned_int = 0;
int zero_int = 0;
unsigned long int zero_unsigned_long_int = 0;
long int zero_long_int = 0;
float zero_float = 0;
double zero_double = 0;
long double zero_long_double = 0;
real zero_real = 0;

APD_PUSH_FUNCTION(4, "from \"%s\", line %d", cfile, cline);

va_start(variable_parameters, is_object);
return_me = _mdarrays_allocate_macro_guts(cfile, cline, type_string, type_size, is_object,
                                          variable_parameters);
va_end(variable_parameters);

// Point value_pointer to the correct type
if (strcmp(type_string, "unsigned char") == 0)
  ffps.value_pointer = (void *)(&zero_unsigned_char);
else if (strcmp(type_string, "char") == 0)
  ffps.value_pointer = (void *)(&zero_char);
else if (strcmp(type_string, "uint") == 0)
  ffps.value_pointer = (void *)(&zero_unsigned_int);
else if (strcmp(type_string, "int") == 0)
  ffps.value_pointer = (void *)(&zero_int);
else if (strcmp(type_string, "unsigned long int") == 0)
  ffps.value_pointer = (void *)(&zero_unsigned_long_int);
else if (strcmp(type_string, "long int") == 0)
  ffps.value_pointer = (void *)(&zero_long_int);
else if (strcmp(type_string, "float") == 0)
  ffps.value_pointer = (void *)(&zero_float);
else if (strcmp(type_string, "double") == 0)
  ffps.value_pointer = (void *)(&zero_double);
else if (strcmp(type_string, "long double") == 0)
  ffps.value_pointer = (void *)(&zero_long_double);
else if (strcmp(type_string, "real") == 0)
  ffps.value_pointer = (void *)(&zero_real);
else APD_DIE("variable type \"%s\" unknown (only basic types can be zeroed", type_string);

// Now do fill
_mdarrays_foreach(cfile, cline, return_me, _macro_fill_foreach_function, &ffps);

APD_POP_FUNCTION();
return return_me;
} // end _mdarrays_allocate_macro_fill()

// ----------------------------------------------------
// - - - - DEALLOCATION - - - -
// ----------------------------------------------------

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_deallocate_recurse(char *cfile, uint cline, void *array, 
                                  mdarray_info_struct *mdarray_info, uint d_counter)
{
char **char_pointer2 = NULL;
int loop;
int current_dimension_size;

current_dimension_size = mdarray_info->sizes[d_counter-1];
APD_PUSH_FUNCTION(4, "dimensions=%d, d_counter=%d, type_size=%d, current size=%d",
      mdarray_info->dimensions, d_counter, mdarray_info->type_size, current_dimension_size);

MDARRAYS_ASSERT(mdarray_info->dimensions > 0,
                "Fatal error: dimensions (%d < 1)", mdarray_info->dimensions);
MDARRAYS_ASSERT(d_counter <= mdarray_info->dimensions);
MDARRAYS_ASSERT(mdarray_info->type_size > 0);
MDARRAYS_ASSERT(mdarray_info->sizes != NULL);

// Get dimension size for this dimension
MDARRAYS_ASSERT(current_dimension_size > 0,
           "negative or zero dimension size (dimension %d)", d_counter);

if (d_counter < mdarray_info->dimensions)
  {
    char_pointer2 = (char **)array;  // Must use a real type to use indexing in for loop below
    for (loop = 0; loop < current_dimension_size; loop++)
      {
        _mdarrays_deallocate_recurse(cfile, cline,
                                     (void *)(char_pointer2[loop]), mdarray_info, d_counter + 1);
      }
    APD_REPORT(5, "calling free($%lX) - #1", (unsigned long int)array);
    free(array);
  } // end if d_counter < dimensions
else
  {
    APD_REPORT(5, "calling free($%lX) - #2", (unsigned long int)array);
    free(array);
  }

APD_POP_FUNCTION();
}

// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++
// This main routine just basically sets up the arguments list to pass to
// _mdarrays_deallocate_recurse()
void _mdarrays_deallocate(char *cfile, uint cline, void * array)
{
char *char_pointer = NULL;
char **char_pointer2 = NULL;
mdarray_info_struct *mdarray_info;
uint loop;

MDARRAYS_ASSERT(array != NULL);
APD_PUSH_FUNCTION(4, "from \"%s\", line %d: deallocating mdarray @%lX", 
                  cfile, cline, (unsigned long int)array);

// Extra info from array header info
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, "unknown");

// Recurse and free if dimensions > 1
if (mdarray_info->dimensions > 1)
  {
    char_pointer2 = (char **)array;  // Must use a real type to use indexing in for loop below
    for (loop = 0; loop < mdarray_info->sizes[0]; loop++)
      {
        _mdarrays_deallocate_recurse(cfile, cline,
                                     (void *)(char_pointer2[loop]), mdarray_info, 2u);
      }
  }
  
// Deallocate top level special due to mdarray_info structure
// Move back to original pointer position and free memory
MDARRAYS_ASSERT(sizeof(char) == 1);
char_pointer = (char *)array;
char_pointer = char_pointer - sizeof(mdarray_info_struct);
mdarray_info_struct *info = (mdarray_info_struct *)char_pointer;
info->tag = 0; // wipe tag so we will know this is no longer an mdarray
APD_REPORT(5, "calling free($%lX) - #0", (unsigned long int)char_pointer);
free(char_pointer);

APD_POP_FUNCTION();
}
