#include <mdarrays.h>
#include <internal.h>

// Save contents of an mdarray or nonmdarray to a file.
// Note that these are not
// mix and match; you can't save a nonmdarray, then load to an
// mdarray, it won't work except for the 1D case.

/* Save the array to file */
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_save_recurse(char *cfile, uint cline, FILE *file, void * array, 
                            uint dimensions, uint d_counter, 
                            uint type, uint type_size, uint *sizes, 
                            print_style_struct *save_style)
{
uint index;
uint current_dimension_size;
long double **double_pointer;
size_t fwrite_result;
void *read_here;

MDARRAYS_ASSERT(file != NULL);
MDARRAYS_ASSERT(array != NULL);

APD_PUSH_FUNCTION(4, "saving array @ %lX (dim=%d)",
	    (unsigned long int)(array), dimensions);
MDARRAYS_ASSERT(dimensions > 0);
MDARRAYS_ASSERT(d_counter <= dimensions);
MDARRAYS_ASSERT(type_size > 0);
MDARRAYS_ASSERT(sizes != NULL);
current_dimension_size = sizes[d_counter-1];

if (d_counter == dimensions)
  {
    // Output here
    for (index = 0; index < current_dimension_size; index++)
      {
        if (type == MDARRAYS_GPOINTER_TYPE)
          {
            read_here = (void *)(&(((char *)array)[index * type_size]));
            APD_REPORT(5, "Writing %d bytes at @%lX", type_size, (unsigned long int)read_here);
            fwrite_result = fwrite(read_here, 1, type_size, file);
            //APD_ACTION(5, fflush(file); ); // Valgrind will catch things faster if this happens
            MDARRAYS_ASSERT(fwrite_result == type_size, 
                              "error writing gpointer array to file.");
          }
        else
          {
            // Print the delimter after all except the last in a row
            bool delimiter_on = !(index == current_dimension_size - 1);
            _mdarrays_print_1D_element(cfile, cline, file, array, 
                                       type, type_size, index, save_style, delimiter_on);
            // Print one value per line for 1D arrays
            //if (dimensions == 1) fprintf(file, "%c", save_style->line_delimiter);
          }
      } // end for index
    if (type != MDARRAYS_GPOINTER_TYPE) 
      {
        fprintf(file, "%c", save_style->line_delimiter);
      }
  } // end if d_counter == dimensions
else
  {
    // Otherwise recurse
    double_pointer = (long double **)array;
    for (index = 0; index < current_dimension_size; index++)
      {
        _mdarrays_save_recurse(cfile, cline, file, double_pointer[index], dimensions, 
                               d_counter+1, type, type_size, sizes, save_style);
      }
  }

APD_POP_FUNCTION();
} // end _mdarrays_save_recurse()

// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_save_guts(char *cfile, uint cline, char *filename, char *description, 
                         void * array, char **tags)
{
FILE *file;
mdarray_info_struct *mdarray_info;

MDARRAYS_ASSERT(filename != NULL);
MDARRAYS_ASSERT(array != NULL);
APD_PUSH_FUNCTION(4, "saving array @%lX to file %s",
      (unsigned long int)array, filename);

// Extra info from array header info
mdarray_info = _mdarrays_get_mdarray_info(cfile, cline, array, "unknown");

// Decide if this will be a direct memory image or not
// If it's a GPOINTER, you're not just pointing to a gpointer,
// you're pointing to a structure or something.
if (mdarray_info->type == MDARRAYS_GPOINTER_TYPE)
  {
    // direct memory to disk - open file as binary
    file = fopen(filename, "wb");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for writing (binary mode)", filename);
  }
else
  {
    file = fopen(filename, "w");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for writing", filename);
  }

if (description != NULL)
  {
    fprintf(file, "#%s\n", description);
  }
else
  {
    fprintf(file, "#\n");
  }

if (tags != NULL)
  {
    _mdarrays_print_tagged(cfile, cline, file, array, tags);
  }
else
  {
    // Recurse and save
    _mdarrays_save_recurse(cfile, cline, file, array, 
                           mdarray_info->dimensions, 1, mdarray_info->type,
                           mdarray_info->type_size, mdarray_info->sizes,
                           &(mdarray_info->save_style));
  }
  
fclose(file);
APD_REPORT(4, "closed %s", filename);
APD_POP_FUNCTION();
} // end mdarrays_save_guts()

// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_save(char *cfile, uint cline, char *filename, char *description, void * array)
{
_mdarrays_save_guts(cfile, cline, filename, description, array, NULL);
}

// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_save_tagged(char *cfile, uint cline, char *filename, char *description, 
                           void * array, char **tags)
{
_mdarrays_save_guts(cfile, cline, filename, description, array, tags);
}

// ++++++++++++++++++++++++++++++++++++++++++++
//
// ++++++++++++++++++++++++++++++++++++++++++++
void _mdarrays_save_nonmdarray(char *cfile, uint cline, 
                               char *filename, void * array, uint type, unsigned
                               int type_size, uint dimensions, uint *sizes,
                               char delimiter, char line_delimiter,
                               uint print_type, uint print_width, uint print_precision)
{
uint col;
uint elements;
void *read_here;
FILE *file;
size_t fwrite_result;
print_style_struct save_style;
save_style.delimiter = delimiter;
save_style.line_delimiter = line_delimiter;
save_style.type = print_type;
save_style.width = print_width;
save_style.precision = print_precision;

APD_PUSH_FUNCTION(4, "saving nonmdarray @ %lX (dim=%d, type=%d, type_size=%d)",
	    (unsigned long int)array, dimensions, type, type_size);

MDARRAYS_ASSERT(dimensions > 0, "Fatal error: dimensions < 1");

if (type == MDARRAYS_GPOINTER_TYPE)
  {
    // direct memory to disk - open file as binary
    file = fopen(filename, "wb");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for writing (binary mode)", filename);
  }
else
  {
    file = fopen(filename, "w");
    MDARRAYS_ASSERT(file != NULL, "error opening file: \"%s\"", filename);
    APD_REPORT(4, "opened %s for writing", filename);
  }

// Get total number of elements in array
elements = 1;
for (col = 0; col < dimensions; col++)
  {
    MDARRAYS_ASSERT(sizes[col] > 0);
    elements = elements * (uint)sizes[col];
  }
APD_REPORT(4, "Total number of array elements: %d\n", elements);

// Print out all elements, in memory order
for (col = 0; col < elements; col++)
  {
    if (type == MDARRAYS_GPOINTER_TYPE)
      {
        read_here = (void *)(&(((char *)array)[col * type_size]));
        fwrite_result = fwrite(read_here, 1, type_size, file);
        MDARRAYS_ASSERT(fwrite_result == type_size, "error writing gpointer array to file.");
      }
    else
      {
        bool delimiter_on = !(col == elements - 1);
        _mdarrays_print_1D_element(cfile, cline, file, array, type, type_size, col, &save_style, 
                                   delimiter_on);
      }
  } // end for col

fclose(file);

APD_POP_FUNCTION();
} // end _mdarrays_save_nonmdarray()
